COLOR_PALETTE = {
    "success": "rgb(113,176,109)",
    "warning": "rgb(176,169,109)",
    "error": "rgb(176,109,109)"
}

STARFLOW_HELM_PATH = 'starflow/starflow'
STARFLOW_HELM_VERSION = '0.1.6'
STARFLOW_HELM_REPOSITORY = 'https://prochain-star-atlas.github.io/starflow/'